import argparse
import subprocess
from pathlib import Path
from .utils import define_python_interpreter
from ..policy import split

BATCH_SCRIPT = """\
#!/bin/bash
#
#SBATCH --job-name=split
#SBATCH --ntasks=1
#SBATCH --cpus-per-task={cpus_per_task}
#SBATCH --partition={partition}
#SBATCH --time={time}
#SBATCH --mem={mem}
#SBATCH --output={output_log}
#SBATCH --error={error_log}

{env_cmd}

split-log -o {output_dir} -l {lines} {log}
"""

def parse_args():
    parser = argparse.ArgumentParser(
        description="Splits a GPFS policy log into multiple parts for batch array processing.",
        formatter_class=argparse.RawTextHelpFormatter
    )
    parser.add_argument('log',
                        type=Path,
                        help="Path to a raw GPFS log file. The log can be either uncompressed or gz compressed.")
    
    parser.add_argument('-o','--output-dir',
                        type=Path,
                        default = None,
                        help="Directory to store the output parquet. The parquet file will have the same name as the input. Defaults to ./input/../chunks")
    parser.add_argument('-l', '--lines', type=int, default=int(5e6),
                        help="Number of lines to split the log file by")

    slurm = parser.add_argument_group(title='Slurm Options')
    slurm.add_argument('--batch', action='store_true', default=False,
                       help="Run as a batch job. Otherwise use the current processing environment")
    slurm.add_argument('-n','--cpus-per-task', type=int, default=24,
                       help="Number of cores assigned to the job. Ntasks is always set to 1")
    slurm.add_argument('-p','--partition', type=str, default='amd-hdr100')
    slurm.add_argument('-t','--time',type=str,default='02:00:00')
    slurm.add_argument('-m','--mem',type=str,default='48G')
    slurm.add_argument('--slurm-log-dir',type=Path,default='./out',
                       help='Output log directory. If the directory does not exist, it will be created automatically')
    args = parser.parse_args()
    return vars(args)

def submit_batch(**kwargs):
    env_cmd = define_python_interpreter()
    kwargs.update({"env_cmd":env_cmd})
    
    kwargs.get('slurm_log_dir').mkdir(exist_ok=True,parents=True,mode = 0o2770)
    output_log = kwargs.get('slurm_log_dir').joinpath('split.out')
    error_log = kwargs.get('slurm_log_dir').joinpath('split.err')
    kwargs.update({'output_log':output_log,'error_log':error_log})

    script = BATCH_SCRIPT.format(**kwargs)

    subprocess.run(['sbatch'],input=script,shell=True,text=True)
    pass

def split_log():
    args = parse_args()

    if args['output_dir'] is None:
        args['output_dir'] = args['log'].parent.parent.joinpath('chunks')
    args['output_dir'].mkdir(exist_ok = True, mode = 0o2770)

    if args.get('batch'):
        submit_batch(**args)
    else:
        split(**args)
    pass